@@ -16,6 +16,7 @@ use Astro::Coord::Precession 'precess';
 use Crypt::JWT qw(encode_jwt decode_jwt);
 use CSS::Inliner;
 use DateTime;
+use DBI;
 use File::ShareDir 'dist_dir';
 use HTML::FormatText;
 use HTML::TreeBuilder;
@@ -26,18 +27,22 @@ use JSON::XS;
 use Math::DCT ':all';
 use Math::MatrixReal;
 use MCE::Loop;
+use SQL::Abstract::Classic;
+use SQL::Inserter;
 use Sys::Info;
 use Sys::Info::Constants qw( :device_cpu );
+use Text::Fuzzy;
+use Text::Levenshtein::Damerau::XS;
 
 use Exporter 'import';
 our @EXPORT = qw(system_identity suite_run calc_scalability);
 our $datadir = dist_dir("Benchmark-DKbench");
 
-our $VERSION = '2.0_003';
+our $VERSION = '2.0_004';
 
 =head1 NAME
 
-Benchmark::DKbench - Perl Benchmark
+Benchmark::DKbench - Perl CPU Benchmark
 
 =head1 SYNOPSIS
 
@@ -61,9 +66,11 @@ performance of systems when running computationally intensive Perl (both pure Pe
 and XS) workloads. Runs single and multi-threaded and can be fully customized to
 the benchmarks that better suit your own scenario.
 
-The only non-CPAN libraries required to install/run the suite are libdb, libjpeg
-(with headers). There are a few prerequisite CPAN libraries, plus an optional one
-(L<BioPerl>, requires libxml on system) which enables extra benchmarks.
+The only non-CPAN software required to install/run the suite is libjpeg (with
+headers) and a build environment (C compiler etc) for the XS/C modules.
+Other than that there are quite a few prerequisite CPAN libraries, as they are used
+for the various benchmarks, plus an optional one (L<BioPerl>, requires libxml on
+system) which enables a couple of extra benchmarks.
 
 =head1 SCRIPTS
 
@@ -74,8 +81,8 @@ benchmarking environment.
 =head2 C<dkbench>
 
 The main script that runs the DKbench benchmark suite. If L<BioPerl> is installed,
-you may want to start your first run with C<dkbench --setup>. Otherwise, there are
-many options to control number of threads, iterations, which benchmarks to run etc:
+you may want to start with C<dkbench --setup>. But beyond that, there are many
+options to control number of threads, iterations, which benchmarks to run etc:
 
  dkbench [options]
 
@@ -95,6 +102,7 @@ many options to control number of threads, iterations, which benchmarks to run e
  --sleep <i>           : Sleep for <i> secs after each benchmark.
  --setup               : Download the Genbank data to enable the BioPerl tests.
  --datapath <path>     : Override the path where the expected benchmark data is found.
+ --ver <num>           : Skip benchmarks added after the specified version.
 
 The default run (no options) will run all the benchmarks both single-threaded and
 multi-threaded (using all detected CPU cores/hyperthreads) and show you scores and
@@ -173,6 +181,10 @@ using L<Crypt::JWT>.
 
 =item * C<DateTime> : Creates and manipulates L<DateTime> objects.
 
+=item * C<DBI/SQL> : Creates a mock L<DBI> connection (using L<DBD::Mock>) and passes
+it insert/select statements using L<SQL::Inserter> and L<SQL::Abstract::Classic>.
+The latter is quite slow at creating the statements, but it is widely used.
+
 =item * C<Digest> : Creates MD5, SH1 and SHA-512 digests of a large string.
 
 =item * C<Encode> : Encodes/decodes large strings from/to UTF-8/16, cp-1252.
@@ -194,9 +206,9 @@ matrices.
 =item * C<Moose> : Creates L<Moose> objects.
 
 =item * C<Moose prove> : Runs 110 tests from the Moose 2.2201 test suite. The least
-CPU-intensive test, most of the time will be spent loading the interpreter and
-the Moose module for each test, which is behaviour representative of how a Perl
-test suite runs by default.
+CPU-intensive test (which is why there is the option C<--no_prove> to disable it),
+most of the time will be spent loading the interpreter and the Moose module for each
+test, which is behaviour representative of how a Perl test suite runs by default.
 
 =item * C<Primes> : Calculates all primes up to 7.5 million. Small number with
 repeat was chosen to keep low memory (this is a pure Perl function no Math libraries).
@@ -209,6 +221,10 @@ repeat was chosen to keep low memory (this is a pure Perl function no Math libra
 a utf8 string. Perl version can make a big difference, as Unicode behaviour has
 changed (old Perl versions are faster but less strict in general).
 
+=item * C<Text::Fuzzy> : The edit distance for strings of various lengths (up to 1000)
+are calculated using L<Text::Fuzzy> and L<Text::Levenshtein::Damerau::XS> (the latter
+is faster for Levenshtein-Damerau over long strings.
+
 =item * C<Time::Piece> : Creates and manipulates/converts Time::Piece objects. It
 is disabled by default on MacOS (unless C<--no_mce> is specified), as it runs
 extremely slow when forked on this platform. In general you may want to skip this
@@ -277,12 +293,13 @@ the same terms as the Perl 5 programming language system itself.
 =cut
 
 sub benchmark_list {
-    return {               # idx : 0 = result, 1 = ref time, 2 = func, 3 = quick test, 4 = normal test
+    return {               # idx : 0 = result, 1 = ref time, 2 = func, 3 = quick test, 4 = normal test, 5 = ver
         'Astro'            => ['e71c7ae08f16fe26aea7cfdb72785873', 5.674, \&bench_astro, 20000, 80000],
         'BioPerl Codons'   => ['97c443c099886ca60e99f7ab9df689b5', 8.752, \&bench_bioperl_codons, 3, 5, 1],
         'BioPerl Monomers' => ['d29ed0a5c205c803c112be1338d1f060', 5.241, \&bench_bioperl_mono, 6, 20],
         'Crypt::JWT'       => ['d41d8cd98f00b204e9800998ecf8427e', 6.451, \&bench_jwt, 250, 900],
         'CSS::Inliner'     => ['82c1b6de9ca0500a48f8a8df0998df3c', 4.603, \&bench_css, 2, 5],
+        'DBI/SQL'          => ['2b8252daad9568a5b39038c696df4be3', 5.700, \&bench_dbi, 5000, 15000, 2.1],
         'DateTime'         => ['b08d2eeb994083b7422f6c9d86fed2c6', 6.198, \&bench_datetime, 5000, 15000],
         'Digest'           => ['4b69f6cf0f53cbf6c3444f2f767dd21d', 4.513, \&bench_digest, 50, 250],
         'Encode'           => ['PASS 1025',                        5.725, \&bench_encode, 40, 120],
@@ -296,7 +313,8 @@ sub benchmark_list {
         'Primes'           => ['4266f70a7a9efb3484cf5d98eba32244', 3.680, \&bench_primes_m, 2, 5],
         'Regex/Subst'      => ['30ce365b25f3d597578b3bdb14aa3f57', 4.652, \&bench_regex_asc, 8, 24],
         'Regex/Subst utf8' => ['857eb4e63a4d174ca4a16fe678f7626f', 5.703, \&bench_regex_utf8, 3, 10],
-        'Time::Piece'      => ['5d393b000a42a626bad7208f8be4121e', 5.907, \&bench_timepiece, 75_000, 275_000],
+        'Text::Fuzzy'      => ['a71709068d1051ec5c2e05df82638dbe', 5.816, \&bench_textfuzzy, 12, 42, 2.1],
+        'Time::Piece'      => ['2d4b149fe7f873a27109fc376d69211b', 5.907, \&bench_timepiece, 75_000, 275_000],
     };
 }
 
@@ -366,7 +384,7 @@ sub calc_scalability {
         my @res2 = min_max_avg($stats2->{$bench}->{times});
         push @perf, $res1[2]/$res2[2]*$threads if $res2[2];
         push @scal, $res1[2]/$res2[2]*100 if $res2[2];
-        print pad_to("$bench:",24).pad_to(sprintf("%.2f",$perf[-1]),24).pad_to(sprintf("%2d",$scal[-1]),24)."\n";
+        print pad_to("$bench:",24).pad_to(sprintf("%.2f",$perf[-1]),24).pad_to(sprintf("%2.0f",$scal[-1]),24)."\n";
     }
     print (("-"x34)."\n");
     my ($avg1) = min_max_avg($stats1->{total}->{$display});
@@ -379,7 +397,7 @@ sub calc_scalability {
     @perf = min_max_avg(\@newperf);
     @scal = min_max_avg(\@newscal);
     print pad_to("Multi/Single perf:").sprintf("%.2fx\t(%.2f - %.2f)", $perf[2], $perf[0], $perf[1])."\n";
-    print pad_to("Multi scalability:").sprintf("%2d%% \t(%d%% - %d%%)", $scal[2], $scal[0], $scal[1])."\n";
+    print pad_to("Multi scalability:").sprintf("%2.1f%% \t(%.0f%% - %.0f%%)", $scal[2], $scal[0], $scal[1])."\n";
 }
 
 sub run_iteration {
@@ -395,6 +413,7 @@ sub run_iteration {
         next if $opt->{skip_prove} && $bench =~ /prove/;
         next if !$opt->{bio_codons} && $bench =~ /Codons/;
         next if $opt->{skip_timep} && $bench =~ /Time::Piece/;
+        next if $opt->{ver} && $benchmarks->{$bench}->[5] && $opt->{ver} < $benchmarks->{$bench}->[5];
         next if $opt->{exclude} && $bench =~ /$opt->{exclude}/;
         next if $opt->{include} && $bench !~ /$opt->{include}/;
         if ($bench =~ /Bio/) {
@@ -548,6 +567,31 @@ sub bench_datetime {
     return $d->hexdigest;
 }
 
+sub bench_dbi {
+    my $iter = shift;
+    my $d    = Digest->new("MD5");
+    my $dbh  = DBI->connect( 'DBI:Mock:', '', '' );
+    my ($data, $cols) = _db_data();
+
+    foreach (1..$iter) {
+        my $inserter = SQL::Inserter->new(
+            dbh    => $dbh,
+            table  => 'table',
+            cols   => $cols,
+            buffer => 2
+        );
+        $inserter->insert($data->[int(rand(20))]) for 1..2;
+        $d->add($dbh->last_insert_id);
+        my $sql = SQL::Abstract::Classic->new();
+        my ($stmt, @bind) = $sql->insert('table', $data->[int(rand(20))]);
+        $d->add($dbh->quote($stmt));
+        ($stmt, @bind) = $sql->select('table', $cols->[int(rand(20))], [map {_rand_where()} 1..int(rand(3)+1)]);
+        $d->add($dbh->quote($stmt._random_str(5)));
+        my $dbh2 = DBI->connect( 'DBI:Mock:', '', '' );
+    }
+    return $d->hexdigest;
+}
+
 sub bench_dct {
     my $iter = shift;
     my $d    = Digest->new("MD5");
@@ -867,6 +911,29 @@ sub bench_subst {
     return "$count Replaced";
 }
 
+sub bench_textfuzzy {
+    my $iter = shift;
+    my $d    = Digest->new("MD5");
+    my $data = _fuzzy_data();
+    my $diff;
+    foreach (1..$iter) {
+        foreach my $sz (qw/10 100 1000/) {
+            my $i = int(rand(10));
+            my $tf = Text::Fuzzy->new($data->{$sz}->[$i]);
+            $diff = $tf->distance($data->{$sz}->[$_]) for 0..9;
+            $d->add($diff);
+            $tf = Text::Fuzzy->new($data->{$sz}->[$i], trans => 1, max => 100);
+            $diff = $tf->distance($data->{$sz}->[$_]) for 0..9;
+            $d->add($diff);
+            $diff = Text::Levenshtein::Damerau::XS::xs_edistance(
+                $data->{$sz}->[$i], $data->{$sz}->[$_]
+            ) for 1..9;
+            $d->add($diff);
+        }
+    }
+    return $d->hexdigest;
+}
+
 sub bench_timepiece {
     my $iter = shift;
     my $t    = Time::Piece::localtime(1692119499);
@@ -877,8 +944,8 @@ sub bench_timepiece {
     for (1..$iter) {
         $t += int(rand(1000)-500)*$day;
         $t += 100000*$day if $t->year < 1970;
-        my $str = $t->strftime("%a, %d %b %Y %H:%M:%S");
-        $t = Time::Piece->strptime($str, "%a, %d %b %Y %H:%M:%S");
+        my $str = $t->strftime("%w, %d %m %Y %H:%M:%S");
+        $t = Time::Piece->strptime($str, "%w, %d %m %Y %H:%M:%S");
         my $jd = $t->julian_day;
         $d->add($str,$jd);
     }
@@ -965,9 +1032,9 @@ sub _read_wiki_files {
 
 sub _random_str {
     my $length = shift || 1;
-    my $str    = "";
-
-    $str .= chr(int(rand(95))+32) for 1..$length;
+    my ($base, $rng) = shift ? (65, 26) : (32, 95);
+    my $str = "";
+    $str .= chr(int(rand($rng))+$base) for 1..$length;
     return $str;
 }
 
@@ -978,6 +1045,42 @@ sub _random_uchar {
     return chr($chr);
 }
 
+sub _fuzzy_data {
+    my %data;
+    push @{$data{10}}, join('', map {_random_uchar()} 1..(8+int(rand(5))))
+        for 0..9;
+    push @{$data{100}}, $data{10}->[$_]x10 for 0..9;
+    push @{$data{1000}}, _random_str(50,1)x20 for 0..9;
+    return \%data;
+}
+
+sub _rand_where {
+    my $p = rand();
+    if ($p > 0.5) {
+        return {foo => rand(10)};
+    } elsif ($p > 0.2) {
+        return {bar => {-in => [int($p*10)..int($p*20)]}};
+    } else {
+        my $op = $p > 0.1 ? '-and' : '-or';
+        my @cond = map {_rand_where()} 1..int(rand(3)+1);
+        return {$op => [@cond]};
+    }
+}
+
+sub _db_data {
+    my (@data, @cols);
+    foreach (1..20) {
+        my $d = {
+        id   => int(rand(10000000)),
+        date => \"NOW()",
+        map {"data".$_ => "foo bar" x int(rand(5)+1)} 1..int(rand(20)+1)
+        };
+        push @data, $d;
+        push @cols, [sort keys %$d];
+    }
+    return \@data, \@cols;
+}
+
 sub compare_obj {
     my ($obj1, $obj2) = @_;
     my $t1 = ref($obj1);
